<div v-if="data.{{{widget.widget_id}}}.text">
  <i class="zmdi zmdi-check-circle" style="color: #13ce66; margin-right: 5px"></i>
  <span style="color: #5a6772">
    <span v-html="data.{{{widget.widget_id}}}.text"></span>
  </span>
</div>
